#include "version.hpp"
#include "config.hpp"
#include "service_runner.hpp"
#include "services.hpp"
#include "servers.hpp"
#include "utils/directories.hpp"
#include "templates.hpp"
#include <boost/filesystem.hpp>

#include <iostream>
#include <string>
#include <vector>

namespace dropshell {

void print_version() {
    std::cout << "dropshell version " << VERSION << std::endl;
    std::cout << "Release date: " << RELEASE_DATE << std::endl;
    std::cout << "Author: " << AUTHOR << std::endl;
    std::cout << "License: " << LICENSE << std::endl;
}


void print_help() {
    std::cout << "Usage: dropshell [OPTIONS] COMMAND [ARGS]" << std::endl;
    std::cout << std::endl;
    std::cout << "A tool for managing server configurations" << std::endl;
    std::cout << std::endl;
    std::cout << "Commands:" << std::endl;
    std::cout << "  help           Show this help message" << std::endl;
    std::cout << "  version        Show version information" << std::endl;
    std::cout << "  init DIR       Set the local directory for all server configs and backups" << std::endl;
    std::cout << std::endl;
    std::cout << "  servers        Summary of all configured servers" << std::endl;
    std::cout << "  servers NAME   Show details for specific server" << std::endl;
    std::cout << "  templates      List all available templates" << std::endl;
    std::cout << std::endl;
    std::cout << "  install SERVER [SERVICE]       Install/Update service(s)." << std::endl;
    std::cout << "  backup SERVER [SERVICE]        Backup service(s)." << std::endl;
    std::cout << "  uninstall SERVER [SERVICE]     Uninstall service(s)." << std::endl;
    std::cout << "  COMMAND SERVER [SERVICE]       Run a custom command on service(s)." << std::endl;
    std::cout << std::endl;
}

} // namespace dropshell


bool parseargs(std::string arg2, std::string arg3, std::string & server_name, std::vector<dropshell::ServiceInfo>& servicelist)
{
    if (arg2.empty()) return false;
    server_name = arg2;

    if (arg3.empty()) {
        servicelist = dropshell::get_server_services_info(server_name);
    } else {
        servicelist.push_back(dropshell::get_service_info(server_name, arg3));
    }

    return true;
}

std::string safearg(int argc, char *argv[], int index)
{
    if (index >= argc) return "";
    return argv[index];
}

int main(int argc, char* argv[]) {
    try {
        dropshell::config *cfg = dropshell::get_global_config();

        // Handle commands
        std::string cmd;
        if (argc > 1) {
            cmd = argv[1];
        }

        // don't load old config if we're initializing
        if (cmd == "init") {
            std::string lcd;
            if (boost::filesystem::exists(dropshell::get_local_dropshell_config_path())) {
                std::cerr << "DropShell is already initialised in " << dropshell::get_local_dropshell_config_path() << std::endl;
                std::cerr << "Please manually delete this old config file and re-run the init command." << std::endl;
                return 1;
            }

            if (argc < 3) {
                std::cerr << "Error: init command requires a directory argument" << std::endl;
                return 1;
            }
            try {
                cfg->init_local_config_directory(argv[2]);
                return 0;
            } catch (const std::exception& e) {
                std::cerr << "Error in init: " << e.what() << std::endl;
                return 1;
            }
        }

        if (cmd == "help" || cmd == "-h" || cmd == "--help") {
            dropshell::print_help();
            return 0;
        }

        if (cmd == "version" || cmd == "-V" || cmd == "--version") {
            dropshell::print_version();
            return 0;
        }

        // silently attempt to load the config file.
        cfg->load_config();

        // auto completion stuff.
        std::set<std::string> commands;
        std::vector<dropshell::ServerInfo> servers = dropshell::get_configured_servers();
        for (const auto& server : servers) 
        {
            std::vector<dropshell::ServiceInfo> services = dropshell::get_server_services_info(server.name);
            for (const auto& service : services) 
                commands.merge(dropshell::get_used_commands(server.name, service.service_name));
        }

        if (cmd == "autocomplete_list_commands") { 
            commands.merge(std::set<std::string>{
                "help","version"
            });
            if (!boost::filesystem::exists(dropshell::get_local_dropshell_config_path()))
                commands.insert("init");
            if (cfg->is_config_set()) 
                commands.merge(std::set<std::string>{
                    "servers","templates"
                });
            
            for (const auto& command : commands) {
                std::cout << command << std::endl;
            }
            return 0;
        }

        if (cmd == "autocomplete_list_servers") {
            if (cfg->is_config_set()) 
            {
                auto servers = dropshell::get_configured_servers();
                for (const auto& server : servers) 
                    std::cout << server.name << std::endl;
            }
            return 0;
        }

        if (cmd == "autocomplete_list_services") {
            if (argc < 3) {
                std::cerr << "Error: autocomplete_list_services requires a server name" << std::endl;
                return 1;
            }
            if (cfg->is_config_set())  {
                auto services = dropshell::get_server_services_info(argv[2]);
                for (const auto& service : services) 
                    std::cout << service.service_name << std::endl;
            }
            return 0;
        }

        // ------------------------------------------------------------
        // from here we require the config file to be loaded.
        if (!cfg->is_config_set()) {   
            std::cerr << "Please run 'dropshell init <path>' to initialise the user directory and create a configuration file." << std::endl;
            return 1;
        }

        // No arguments provided
        if (argc < 2) {
            dropshell::print_help();
            return 0;
        }

        if (cmd == "servers") {
            if (argc > 2) {
                // Show details for specific server
                dropshell::show_server_details(argv[2]);
            } else {
                // List all servers
                dropshell::list_servers();
            }
            return 0;
        }

        if (cmd == "templates") {
            dropshell::list_templates();
            return 0;
        }
        
        // handle running a command. 
        for (const auto& command : commands) {
            if (cmd == command) {
                std::string server_name;
                std::vector<dropshell::ServiceInfo> servicelist;
                if (!parseargs(safearg(argc, argv, 2), safearg(argc, argv, 3), server_name, servicelist)) {
                    std::cerr << "Error: " << command << " command requires server name and optionally service name" << std::endl;
                    return 1;
                }

                for (const auto& service_info : servicelist) {
                    dropshell::service_runner service;
                    if (!service.init(server_name, service_info.service_name)) {
                        std::cerr << "Error: Failed to initialize service" << std::endl;
                        return 1;
                    }
                    if (!service.run_command(command)) {
                        std::cerr << command +" failed." << std::endl;
                        return 1;
                    }
                }
                return 0;
            }
        }

        // Unknown command
        std::cerr << "Error: Unknown command '" << cmd << "'" << std::endl;
        dropshell::print_help();
        return 1;

    } catch (const std::exception& e) {
        std::cerr << "Error: " << e.what() << std::endl;
        return 1;
    }
} 